// see Docs/fog_calculation.txt for further information

// tuneable parameters

#define FOG_DENSITY 0.95
#define FOG_MIN_VIEWING_DISTANCE 100
#define FOG_MAX_VIEWING_DISTANCE 3000

#define FOG_GROUNDFOG_BASE_HEIGHT 0
#define FOG_GROUNDFOG_VIEWING_DISTANCE 20
// ground fog starts to appear at this density
#define FOG_GROUNDFOG_KICKIN 0.85
#define FOG_GROUNDFOG_DENSITY \
    (FOG_DENSITY > FOG_GROUNDFOG_KICKIN ? \
        pow((FOG_DENSITY - FOG_GROUNDFOG_KICKIN) / (1.f - FOG_GROUNDFOG_KICKIN), 2) : 0.f)

#define FOG_LAYER0_BASE_HEIGHT 0
#define FOG_LAYER0_HEIGHT1 10
#define FOG_LAYER0_VIEWING_DISTANCE  \
    (FOG_MIN_VIEWING_DISTANCE + FOG_MAX_VIEWING_DISTANCE - (FOG_MAX_VIEWING_DISTANCE * FOG_DENSITY))

#define FOG_LAYER1_BASE_HEIGHT 0
#define FOG_LAYER1_HEIGHT1 100
#define FOG_LAYER1_VIEWING_DISTANCE FOG_MAX_VIEWING_DISTANCE

/////////////////////////////////////////////////////////////////////

#define FOG_ZERO_VISIBILITY_VALUE \
    (1.f / 256.f)

#define FOG_LAYER_FALLOFF(baseHeight, height1) \
    pow(0.5, 1.f / ((height1 - baseHeight) / 100.f))
#define FOG_LAYER_DENSITY(viewingDistance) \
    (log(FOG_ZERO_VISIBILITY_VALUE) / (log(0.5) * (viewingDistance / 100.f)))

#define FOG_DENSITY_SCALE(height, falloff) \
    pow(falloff, (height))
#define FOG_DENSITY_SCALE_ANTIDERIVATIVE(height, falloff) \
    (pow(falloff, (height)) / log(falloff))
#define FOG_FALLOFF_LINEAR(baseHeight, layerTop) \
    (1.f / (layerTop - baseHeight))


// FIXME: the case of camera or target beeing below baseHeight is not handled

//FIXME copied from intercept() in underwater.h
float3 intersect(float3 lineP,
               float3 lineN,
               float3 planeN,
               float  planeD)
{
    float distance = (planeD - dot(planeN, lineP)) / dot(lineN, planeN); //FIXME possible division through 0 ?
    return lineP + lineN * distance;
}

float calcFogLayerValueLinear(
    float baseHeight,
    float layerTop,
    float falloff,
    float density,
    float3 cameraPos,
    float3 worldPos)
{
    float3 minPos = cameraPos.z < worldPos.z ? cameraPos : worldPos;
    float3 maxPos = cameraPos.z < worldPos.z ? worldPos : cameraPos;

    // only the distance below the fog layer top is relevant
    maxPos = maxPos.z > layerTop ?
            intersect(minPos, maxPos - minPos, float3(0, 0, 1), layerTop) :
            maxPos;
    minPos = minPos.z > layerTop ? maxPos : minPos;

    float minDensityScale = 1.f - (max(0, minPos.z - baseHeight) * falloff);
    float maxDensityScale = 1.f - (max(0, maxPos.z - baseHeight) * falloff);
    minDensityScale = max(0, minDensityScale);
    maxDensityScale = max(0, maxDensityScale);

    float avgDensityScale = (minDensityScale + maxDensityScale) / 2.f;

    return (length(maxPos - minPos) / 10000.f) * avgDensityScale * density;
}

float calcFogLayerValue(
    float baseHeight,
    float falloff,
    float density,
    float3 cameraPos,
    float3 worldPos,
    float distance)
{
    // FIXME: this function sometimes calculates erroneous values for avgDensityScale -
    // (verified by setting avgDensityScale to a constant value)
    // this manifests itself as one horizontal line of ramdomly colored pixels on screen
    // occurs seldomly and most likely when a few meters above the water line (altitude 0)

    float minHeight = min(cameraPos.z, worldPos.z) / 10000.f;
    float maxHeight = max(cameraPos.z, worldPos.z) / 10000.f;

    float deltaHeight = maxHeight - minHeight;

    float avgDensityScale = deltaHeight != 0.f ?
        (FOG_DENSITY_SCALE_ANTIDERIVATIVE(maxHeight - baseHeight, falloff) -
            FOG_DENSITY_SCALE_ANTIDERIVATIVE(minHeight - baseHeight, falloff)) / deltaHeight :
        FOG_DENSITY_SCALE(minHeight - baseHeight, falloff);

    return distance * avgDensityScale * density;
}

float calcFogValue(float3 cameraPos, float3 worldPos)
{
    // absolute density = relative density * distance
    float absoluteDensity = 0;

    float distance = length(cameraPos - worldPos) / 10000.f;

    absoluteDensity += calcFogLayerValue(
            FOG_LAYER0_BASE_HEIGHT,
            FOG_LAYER_FALLOFF(FOG_LAYER0_BASE_HEIGHT / 100.f, FOG_LAYER0_HEIGHT1),
            FOG_LAYER_DENSITY(FOG_LAYER0_VIEWING_DISTANCE),
            cameraPos.xyz,
            worldPos.xyz,
            distance);

    absoluteDensity += calcFogLayerValue(
            FOG_LAYER1_BASE_HEIGHT,
            FOG_LAYER_FALLOFF(FOG_LAYER1_BASE_HEIGHT / 100.f, FOG_LAYER1_HEIGHT1),
            FOG_LAYER_DENSITY(FOG_LAYER1_VIEWING_DISTANCE),
            cameraPos.xyz,
            worldPos.xyz,
            distance);

    // ground fog

#define NUM_GROUND_FOG_LAYERS 5
#define ADD_LAYER(layerTop) \
    absoluteDensity += calcFogLayerValueLinear( \
            FOG_GROUNDFOG_BASE_HEIGHT, \
            layerTop, \
            FOG_FALLOFF_LINEAR(FOG_GROUNDFOG_BASE_HEIGHT, layerTop), \
            (FOG_GROUNDFOG_DENSITY * FOG_LAYER_DENSITY(FOG_GROUNDFOG_VIEWING_DISTANCE) / NUM_GROUND_FOG_LAYERS), \
            cameraPos, \
            worldPos);

    ADD_LAYER(100);
    ADD_LAYER(125);
    ADD_LAYER(150);
    ADD_LAYER(175);
    ADD_LAYER(200);
#undef ADD_LAYER
#undef NUM_GROUND_FOG_LAYERS

    return 1.f - pow(0.5, absoluteDensity);
}

float3 applyFog(float3 inputColour, float3 fogColour, float fogValue, bool isSky)
{
    float brightness = isSky ? 1.f : fogValue;
    return shLerp (inputColour, fogColour * brightness, fogValue);
}
